#include "StdAfx.h"
#include <math.h>
#include "camera.h"

double CVector3::GetNorm() const
{
	double retVal;
	retVal = sqrt(X*X + Y*Y + Z*Z);
	return retVal;
}
void CVector3::Normalize()
{
	double length = this-> GetNorm();
	X /= length;
	Y /= length;
	Z /= length;
}
CVector3 CVector3::operator + ( CVector3& v )
{
	CVector3 rVector;
	rVector.SetVector( X + v.X, Y + v.Y, Z + v.Z );
	return rVector;
}
CVector3 CVector3::operator - ( CVector3& v )
{
	CVector3 rVector;
	rVector.SetVector( X - v.X, Y - v.Y, Z - v.Z );
	return rVector;
}
CVector3 CVector3::operator ^ ( CVector3& v ) 
{
	CVector3 retVal;
	retVal.X = Y * v.Z - Z * v.Y;
	retVal.Y = Z * v.X - X * v.Z;
	retVal.Z = X * v.Y - Y * v.X;
	return retVal;
}
double&	CVector3::operator[] ( int idx )
{
	switch( idx )
	{
	default:
	case 0:
		return X;
		break;
	case 1:
		return Y;
		break;
	case 2:
		return Z;
		break;
	}
}

CCamera::CCamera(void)
: m_fTrackballSize(0.4f), m_fZoomScale(0.008f)
{
    m_bCapture = FALSE;
    m_bZoom = FALSE;

    ResetCamera();
}

CCamera::~CCamera(void)
{
}


void CCamera::ResetCamera()
{
    for(int i=0 ; i<4 ; ++i)
	for(int j=0 ; j<4 ; ++j)
	{
	    if(i==j)
			m_matTrackBall[i][j] = 1;
	    else
			m_matTrackBall[i][j] = 0;
	}

    glLoadMatrixf((GLfloat*)m_matTrackBall);

    m_vEyeDirection.SetVector(0,0,1);

    m_fDistance = 20.0f;
}


void CCamera::BeginDrag(const CPoint& _point)
{
    m_bCapture = TRUE;
    m_vOldPoint = TrackBallMapping(_point);
}

void CCamera::EndDrag()
{
    m_bCapture = FALSE;
}

void CCamera::BeginZoom(const CPoint& _point)
{
    m_bZoom = TRUE;
    m_vOldPoint.X = (float)_point.x;
    m_vOldPoint.Y = (float)_point.y;
}
void CCamera::EndZoom()
{
    m_bZoom = FALSE;
}

BOOL CCamera::MouseMove(const CPoint& _point)
{
    if(m_bCapture)
    {
		CVector3 vCurrent = TrackBallMapping(_point);
		CVector3 vDir = vCurrent-m_vOldPoint;
		float fVelocity = vDir.GetNorm();
		if(fVelocity > 0.0001f)
		{
			glPushMatrix();
			CVector3 vRotAxis = m_vOldPoint ^ vCurrent;
			float fRotAngle = fVelocity / (2.0f * m_fTrackballSize) * 90.0f;

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();

			glRotatef(fRotAngle, vRotAxis.X, vRotAxis.Y, vRotAxis.Z);
			
			glMultMatrixf( (GLfloat *) m_matTrackBall );  

			glGetFloatv( GL_MODELVIEW_MATRIX, (GLfloat *) m_matTrackBall );

			m_vOldPoint = vCurrent;

			// compute EyeDirection
			CVector3 vInitial;
			vInitial.SetVector(0,0,1);

			for(int i=0 ; i<3 ; ++i)
			{
				m_vEyeDirection[i] = 0;
				for(int j=0 ; j<3 ; ++j)
					m_vEyeDirection[i] += vInitial[j] * m_matTrackBall[i][j];
			}

			glPopMatrix();
			return TRUE;
		}
    }
    else if(m_bZoom)
    {
		glMatrixMode( GL_PROJECTION );

		float fPixelDiff = _point.x - m_vOldPoint.X; 
		float fZoomFactor = 1.0f + fPixelDiff * m_fZoomScale;

		if(m_fDistance * fZoomFactor < 2.0f)
			fZoomFactor = 2.0f/m_fDistance;

		glScalef( fZoomFactor, fZoomFactor, fZoomFactor );

		m_fDistance *= fZoomFactor;
		m_vOldPoint.X = (float)_point.x;
		m_vOldPoint.Y = (float)_point.y;

		glMatrixMode( GL_MODELVIEW );
		return TRUE;
    }
    return FALSE;
}

CVector3 CCamera::TrackBallMapping(const CPoint& _point)
{
    CVector3 v;
    float d;

    v.X = (2.0f*_point.x - m_szWindowSize.cx) / m_szWindowSize.cx;
    v.Y = (m_szWindowSize.cy - 2.0f*_point.y) / m_szWindowSize.cy;
    v.Z = 0.0;
    d = v.GetNorm();

    d = (d<1.0) ? d : 1.0f;
    v.Z = (float)sqrtf(1.001f - d*d);

    v.Normalize();

    return v;
}